"""Plugin's filters and commands definitions."""

import io
import os

import simplebot
from cachelib import FileSystemCache
from deltachat import Message
from emoji import get_emoji_regexp
from simplebot.bot import DeltaBot, Replies

from . import telegram
from .signal import SignalStickers
from .util import getdefault, sizeof_fmt, upload

DEF_MAX_PACK_SIZE = str(1024 ** 2 * 15)
signal = SignalStickers()


@simplebot.hookimpl
def deltabot_init(bot: DeltaBot) -> None:
    getdefault(bot, "cloud", "https://0x0.st/")
    getdefault(bot, "max_size", DEF_MAX_PACK_SIZE)


@simplebot.hookimpl
def deltabot_start(bot: DeltaBot) -> None:
    path = os.path.join(os.path.dirname(bot.account.db_path), __name__)
    if not os.path.exists(path):
        os.makedirs(path)
    signal.cache = FileSystemCache(
        path, threshold=5000, default_timeout=60 * 60 * 24 * 60
    )


@simplebot.filter
def filter_messages(bot: DeltaBot, message: Message, replies: Replies) -> None:
    """Send me an image or Telegram animated sticker and I will convert it to a Delta Chat sticker for you.

    Send me an emoji to get an sticker representing that emoji.

    Send me a text to search for sticker packs matching that text.

    Also, you can send me an URL of a Signal sticker pack, and I will send you the pack, for example, something like:
    sgnl://addstickers/?pack_id=59d338...&pack_key=56af35...
    """
    if message.chat.is_group():
        return

    if message.filemime.startswith("image/"):
        replies.add(filename=message.filename, viewtype="sticker")
    elif telegram.is_sticker(message.filename):
        replies.add(filename=telegram.convert(message.filename), viewtype="sticker")
    elif signal.is_pack(message.text):
        _process_signal_pack(bot, message, replies)
    elif get_emoji_regexp().fullmatch(message.text):
        pack_url, sticker = signal.get_random_sticker(message.text)
        if pack_url:
            replies.add(
                filename=f"{message.text}.webp",
                bytefile=io.BytesIO(sticker),
                viewtype="sticker",
            )
            replies.add(text=pack_url)
        else:
            replies.add(text=f"❌ No sticker found for: {message.text!r}")
    elif message.text:
        html = signal.search_html(bot.self_contact.addr, message.text)
        if html:
            replies.add(text=f"Results for: {message.text!r}", html=html)
        else:
            replies.add(text=f"❌ No results for: {message.text!r}")


@simplebot.command
def info(payload: str, message: Message, replies: Replies) -> None:
    """Get pack info.

    Example:
    /info sgnl://addstickers/?pack_id=59d338...&pack_key=56af35...
    """
    if signal.is_pack(payload):
        text, cover = signal.get_pack_metadata(payload)
        replies.add(
            text=text, filename="cover.webp", bytefile=io.BytesIO(cover), quote=message
        )
    else:
        replies.add("❌ Unknow pack URL", quote=message)


def _process_signal_pack(bot: DeltaBot, message: Message, replies: Replies) -> None:
    title, path = signal.download_pack(bot.account.get_blobdir(), message.text)
    size = os.stat(path).st_size
    if size > int(getdefault(bot, "max_size", DEF_MAX_PACK_SIZE)):
        url = upload(bot, path)
        if url:
            replies.add(
                text=f"Name: {title}\nSize: {sizeof_fmt(size)}\nDownload: {url}",
                quote=message,
            )
        else:
            replies.add(text=f"❌ Pack too big ({sizeof_fmt(size)})", quote=message)
        os.remove(path)
    else:
        replies.add(filename=path, quote=message)



class TestPlugin:
    def test_filter(self, mocker) -> None:
        msg = mocker.get_one_reply(filename="image.png")
        assert msg.filename
        assert msg.is_sticker()

        msg = mocker.get_one_reply(
            "https://signal.art/addstickers/#pack_id=59d3387717104e38a67f838e7ad0208c&pack_key=56af35841874d6fe82fa2085e8e5ed74dba5d187af007d3b4d8a3711dd722ad7"
        )
        assert msg.filename.endswith("zip")

        msg = mocker.get_one_reply("cat")
        assert msg.has_html()

        assert mocker.get_replies("❌")  # random behavior



"""Signal sticker pack downlad tools"""

import os
import random
import zipfile
from tempfile import NamedTemporaryFile
from urllib import parse
from urllib.parse import quote, quote_plus

import anyio
from cachelib import BaseCache, NullCache
from emoji import demojize
from signalstickers_client import StickersClient
from signalstickers_client.models.sticker_pack import StickerPack

from .util import session


class SignalStickers:
    """Class to manage Signal stickers and packs."""

    def __init__(self, cache: BaseCache = NullCache(), max_search_results=100) -> None:
        self.cache = cache
        self.max_search_results = max_search_results

    def _get_cached_pack(self, pack_id: str, pack_key: str) -> StickerPack:
        pack = self.cache.get(pack_id)
        if not pack:
            pack = anyio.run(_get_pack, pack_id, pack_key)
            self.cache.set(pack_id, pack)
        return pack

    def _get_cached_sticker(
        self, sticker_id: int, pack_id: str, pack_key: str
    ) -> bytes:
        cache_key = f"{pack_id}+{sticker_id}"
        sticker = self.cache.get(cache_key)
        if not sticker:
            sticker = anyio.run(_get_sticker, sticker_id, pack_id, pack_key)
            self.cache.set(cache_key, sticker)
        return sticker

    def _get_metadata(self) -> list:
        url = "https://api.signalstickers.com/v1/packs/"
        data = self.cache.get(url)
        if not data:
            with session.get(url) as resp:
                resp.raise_for_status()
                data = resp.json()
            self.cache.set(url, data, timeout=60 * 60 * 24 * 3)
        return data

    @staticmethod
    def is_pack(url: str) -> bool:
        return url.startswith(
            ("https://signal.art/addstickers/", "sgnl://addstickers/")
        )

    def get_pack_metadata(self, pack_url: str) -> tuple:
        pack = self._get_cached_pack(*_parse_url(pack_url))
        text = f"Title: {pack.title or 'NO TITLE'}\nAuthor: {pack.author or 'ANONYMOUS'}\nStickers: {len(pack.stickers)}"
        for mpack in self._get_metadata():
            if mpack["meta"]["id"] == pack.id:
                tags = _get_tags(mpack)
                if tags:
                    text += f"\nTags: {', '.join(tags)}"
                break
        return text, self._get_cached_sticker(pack.cover.id, pack.id, pack.key)

    def download_pack(self, base_dir: str, pack_url: str) -> tuple:
        pack = self._get_cached_pack(*_parse_url(pack_url))
        pack_name = quote(pack.title, safe="")
        with NamedTemporaryFile(
            dir=base_dir,
            prefix=pack_name + "-",
            suffix=".zip",
            delete=False,
        ) as file:
            zip_path = file.name
        with zipfile.ZipFile(zip_path, "w", compression=zipfile.ZIP_DEFLATED) as zfile:
            for sticker in pack.stickers:
                desc = demojize(sticker.emoji, delimiters=("", ""))
                name = os.path.join(
                    pack_name, f"{sticker.id}.{desc}+{sticker.emoji}.webp"
                )
                data = self._get_cached_sticker(sticker.id, pack.id, pack.key)
                zfile.writestr(name, data)
        return pack.title, zip_path

    def search(self, query: str) -> list:
        query = query.lower()
        data = self._get_metadata()
        packs = []
        count = 0
        for pack in data:
            title = pack["manifest"].get("title", "").lower()
            if query in title or query in _get_tags(pack):
                packs.append(pack)
                count += 1
                if count == self.max_search_results:
                    break
        return packs

    def search_html(self, addr: str, query: str) -> str:
        html = ""
        for pack in self.search(query):
            url = _get_pack_url(pack["meta"]["id"], pack["meta"]["key"])
            more_url = f"mailto:{addr}?body=/info+{quote_plus(url)}"
            url = f"mailto:{addr}?body={quote_plus(url)}"
            title = pack["manifest"].get("title", "NO TITLE")
            author = pack["manifest"].get("author", "ANONYMOUS")
            tags = ", ".join(_get_tags(pack))
            html += f'Title: {title}<br/>Author: {author}<br/>Tags: {tags}<br/>\n<a href="{more_url}">MORE</a> | <a href="{url}">DOWNLOAD</a><br/><hr/>'
        return html

    def get_random_sticker(self, emoji: str) -> tuple:
        data = self._get_metadata()
        i = 0
        while i < 200 and data:
            mpack = data.pop(random.randrange(len(data)))
            pack = self._get_cached_pack(mpack["meta"]["id"], mpack["meta"]["key"])
            for sticker in pack.stickers:
                if sticker.emoji == emoji:
                    url = _get_pack_url(pack.id, pack.key)
                    return url, self._get_cached_sticker(sticker.id, pack.id, pack.key)
            i += 1
        return "", b""


async def _get_pack(pack_id: str, pack_key: str) -> StickerPack:
    async with StickersClient() as client:
        return await client.get_pack_metadata(pack_id, pack_key)


async def _get_sticker(sticker_id: int, pack_id: str, pack_key: str) -> bytes:
    async with StickersClient() as client:
        return await client.download_sticker(sticker_id, pack_id, pack_key)


def _parse_url(url: str) -> tuple:
    result = parse.urlsplit(url)
    if result.query:
        data = dict(parse.parse_qsl(result.query))
    else:
        data = dict(parse.parse_qsl(result.fragment))
    return data["pack_id"], data["pack_key"]


def _get_pack_url(pack_id: str, pack_key: str) -> str:
    return f"sgnl://addstickers/?pack_id={pack_id}&pack_key={pack_key}"


def _get_tags(mpack: dict) -> set:
    tags = set(tag.lower() for tag in mpack["meta"].get("tags", []))
    if mpack["meta"].get("nsfw", False):
        tags.add("nsfw")
    if mpack["meta"].get("animated", False):
        tags.add("animated")
    if mpack["meta"].get("original", False):
        tags.add("original")
    return set(tags)



"""Setup module installation."""

import os

from setuptools import find_packages, setup


def load_requirements(path: str) -> list:
    """Load requirements from the given relative path."""
    with open(path, encoding="utf-8") as file:
        requirements = []
        for line in file.read().split("\n"):
            if line.startswith("-r"):
                dirname = os.path.dirname(path)
                filename = line.split(maxsplit=1)[1]
                requirements.extend(load_requirements(os.path.join(dirname, filename)))
            elif line and not line.startswith("#"):
                requirements.append(line.replace("==", ">="))
        return requirements


if __name__ == "__main__":
    MODULE_NAME = "simplebot_stickers"
    DESC = "Sticker packs for all your Delta Chat needs (SimpleBot plugin)"

    with open("README.rst", encoding="utf-8") as fh:
        long_description = fh.read()

    setup(
        name=MODULE_NAME,
        setup_requires=["setuptools_scm"],
        use_scm_version={
            "root": ".",
            "relative_to": __file__,
            "tag_regex": r"^(?P<prefix>v)?(?P<version>[^\+]+)(?P<suffix>.*)?$",
            "git_describe_command": "git describe --dirty --tags --long --match v*.*.*",
        },
        description=DESC,
        long_description=long_description,
        long_description_content_type="text/x-rst",
        author="adbenitez",
        author_email="adbenitez@nauta.cu",
        url=f"https://github.com/adbenitez/{MODULE_NAME}",
        keywords="simplebot plugin deltachat stickers",
        license="MPL",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Environment :: Plugins",
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
            "Operating System :: OS Independent",
            "Topic :: Utilities",
        ],
        zip_safe=False,
        include_package_data=True,
        packages=find_packages(),
        install_requires=load_requirements("requirements/requirements.txt"),
        extras_require={
            "test": load_requirements("requirements/requirements-test.txt"),
            "dev": load_requirements("requirements/requirements-dev.txt"),
        },
        entry_points={
            "simplebot.plugins": f"{MODULE_NAME} = {MODULE_NAME}",
        },
    )



"""Module utilities"""

import functools

import requests
from simplebot.bot import DeltaBot

session = requests.Session()
session.headers.update(
    {
        "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:92.0) Gecko/20100101 Firefox/92.0"
    }
)
session.request = functools.partial(session.request, timeout=15)  # type: ignore


def sizeof_fmt(num: float) -> str:
    suffix = "B"
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)  # noqa
        num /= 1024.0
    return "%.1f%s%s" % (num, "Yi", suffix)  # noqa


def upload(bot: DeltaBot, path: str) -> str:
    url = getdefault(bot, "cloud")
    if url:
        with open(path, "rb") as file:
            with session.post(url, files=dict(file=file)) as resp:
                if 200 <= resp.status_code <= 300:
                    return resp.text.strip()
    return ""


def getdefault(bot: DeltaBot, key: str, value: str = None) -> str:
    val = bot.get(key, scope=__name__)
    if val is None and value is not None:
        bot.set(key, value, scope=__name__)
        val = value
    return val



"""Telegram sticker tools"""

import os
from tempfile import NamedTemporaryFile

from lottie.exporters import exporters
from lottie.importers import importers
from lottie.utils.stripper import float_strip


def convert(path: str) -> str:
    prefix = os.path.join(os.path.dirname(path), "sticker")
    with NamedTemporaryFile(prefix=prefix, suffix=".webp", delete=False) as file:
        outpath = file.name
    anim = importers["lottie"].process(path)
    float_strip(anim)
    exporters["webp"].process(
        anim, outpath, lossless=False, method=0, quality=40, skip_frames=10
    )
    return outpath


def is_sticker(path: str) -> bool:
    return path.endswith((".json", ".tgs"))
